# LinkAhead Sample Management Module

This module contains the custom code for sample management like it is
used in GEOMAR's [Biosample Information System
BIS](https://biosamples.geomar.de/).

## Features
### Sample Registration
Samples have unique identifiers (Sample ID) that can be used to persistently
identify them. The Sample ID can for example be printed as QR Code and used to
label the physical samples.

Since online access to LinkAhead might not be available during sampling,
Samples can be registered in advance in order to generate the Sample IDs. This
allows to print labels for future samples and also allows to download a CSV
template that can be filled with information about the sample for later import
into LinkAhead.

Registration is available via the menu in the webinterface "Sample
Management" -> "Register new Sample"

The CSV Template that is generated always has a set of required columns and
possibly additional columns that can be selected during the registration. The
set of columns can be stored for later use via the `RegisterSampleTemplates`.
Such a template can be created using the "Create sample template" menu entry.

### Sample Update
Via the menu in the webinterface "Sample Management" -> "Upload Sample template"
you can upload a CSV template that was created via the registration detailed
above (or via the export, see below). This updates existing Sample Records
according to the data contained in the CSV file. This upload possibly creates
new Records like Events which are related to the updated Samples.

**Important**: Note, that this will override the information of updated samples
in the server and thus can possibly cause data loss if changes in LinkAhead are
not reflected in the uploaded file.

### Child Samples
A Sample can reference another Sample as `Parent sample` to indicate, that the
sample originated form the other one.  If a sample has such a reference, we
call it a child sample and the referenced sample a parent sample.
A lot of information is the same
for both child and parent Sample and this information is not duplicated but it
is only referenced by the parent Sample.

This implies that child samples are treated differently thatn parent samples
during the upload of CSV templates: Some information that is obligatory for
parent sample are not needed for child samples. There for you can create a
different CSV template for child samples via the "Register new child sample"
menu entry.

### Container Registration
Containers have a unique identifier (Container ID) that can be used to
persistently identify them. The Container ID can for example be printed as QR
Code and used to label physical containers.

The "Register new Container" menu entry under "Storage Management" allows to
create new Container Records with the respective IDs. In this process also
information on Container Type, Size and parent Containers are incorporated.

### Container Update
You can upload filled CSV templates to update the information about Containers
stored in LinkAhead

### Container Export
You can create a filled CSV template with the information about Containers as
it is currently stored in LinkAhead.

### Extensibility
The registration, update and export does not only work with RecordTypes and
Properties that are defined within this module, but the data model can be
extended by users of the system: Any Property that is added to the Sample
RecordType can also be used within the workflow. Once an additional Property
was added to the RecordType (say "width"), you can select a corresponding
column during registration with the same name and also upload CSV templates
that have this column. The Sample records will be created or updated to also
have this ("width") column.

If the Property that is added is a RecordType, it must be kind of an enum, i.e.
it should be uniquely identified either by its name or by a sole property.

## Usage

**TODO**: Add explanations, screenshots, ec. to explain the sample
management from a **user** perspective. There is an upload of a
presentation showing the usage in the context of BIS on Zenodo:
https://doi.org/10.5281/zenodo.14056085.

## Setup

Add the `sample-management-custom` directory to the list of custom
directories in your LinkAhead profile.yml. You'll need to set up
[server-side
scripting](https://docs.indiscale.com/caosdb-deploy/server_side_scripting.html)
and give the permissions (`SCRIPTING:EXECUTE:<script_name>`) for the
scripts in `sample-management-custom/caosdb-server/scripting/bin/` to
the desired users and groups. You may also want to add a custom
authentication token configuration so that the registration and or
upload scripts are executed with special permissions.

Add the data model from
`models_and_helper_scripts/sample-management-model.yml` to your
instance's data model (e.g., by [synchronizing with the LinkAhead
advanced user
tools](https://docs.indiscale.com//caosdb-advanced-user-tools/yaml_interface.html)).

## Customization

The sample management module can be customized by overwriting
indivdual code and config files by using files of the same name and
with the same relative path in your LinkAhead profile's custom
directory.  E.g., if you want to overwrite how sample entities are post-processed in the sample upload, you overwrite

```
sample-management-custom/caosdb-server/scripting/bin/sample_helpers/sample_upload_post_processing.py
```

from this repository by a

```
profile/custom/caosdb-server/scripting/bin/sample_helpers/sample_upload_post_processing.py
```

in your server profile. When doing so, it is important that in your
custom file, you use the exact same function names, signatures, and
return types as in the original file. In our example, the custom
`sample_upload_post_processing.py` needs to contain a
`post_process_samples` that takes the samples and the csv data as
arguments and returns the list of post processed samples.


## Testing

There is a pre-configured `test-profile` that you can use for testing
this repository.


## Data Model
This module requires a certain data model in order to work. However, the exact
names of RecordTypes and Properties can be configured. Below, we will use the
default names but also reference the variable names of `constants.yml` that
allow to change those.

### Sample (`variable missing`)
This is the central RecordType for the sample management and used to represent
samples of various types

#### `Parent_sample` (`$parent_sample_prop`)
references the sample from which the sample at hand originated (for example by
cutting the parent sample into multiple pieces0

#### `Main User`
references the person that mainly works with this sample

#### `PI`
references the person that is the responsible principle investigator (PI)

### Event (`$event_rt`)
A sampling event. Each sample reference the event where it was created.

#### Position (`var missing`)
The position (e.g. of an Event) that is defined via longituted, latitude and
possibly elevation
#### "Biome"
#### "Campaign"
#### "Device"
#### "EventType"
#### "igsn_doi_prop"
#### "level"
#### "locality_description_prop"
#### "locality_name_prop"
#### "Sphere"

### Person (`variable missing`)
a natural person

#### `first_name`  (`$first_name_prop`)
#### `last_name`  (`$last_name_prop`)
#### `abbreviation`  (`$abbreviation_prop`)

### `Container` (`$container_rt`)
a physical container that contains Samples and/or other Containers

####`Child_Container`  (`$child_container_prop`)
references Containers that are contained in the Container at hand

#### `container_label`  (`$container_label_prop`)
#### `Container size`  (`$container_size_prop`)
#### `ContainerType`  (`$containertype_rt`)
#### `custom_label` (`$custom_label_prop`)

#### `Responsible_Person` (`$responsible_person_event`)
references the person that is responsilbe for this sample, e.g. the PI

### `RegisterSampleTemplates` (missing var)

## Further Reading

Please refer to the [official documentation](https://docs.indiscale.com/caosdb-deploy/) for more information.

## Contributing

Thank you very much to all contributers—[past, present](https://gitlab.com/linkahead/linkahead-meta/-/blob/dev/HUMANS.md), and prospective ones.

### Code of Conduct

By participating, you are expected to uphold our [Code of Conduct](https://gitlab.com/linkahead/linkahead-meta/-/blob/dev/CODE_OF_CONDUCT.md).

### How to Contribute

* You found a bug, have a question, or want to request a feature? Please
[create an issue](https://gitlab.indiscale.com/caosdb/src/linkahead-sample-management/-/issues).
* You want to contribute code? Please fork the repository and create a merge
request in GitLab and choose this repository as target. Make sure to select
"Allow commits from members who can merge the target branch" under Contribution
when creating the merge request. This allows our team to work with you on your request.
- You can also contact us at **info (AT) indiscale.com**  and join the
  LinkAhead community on
  [#linkahead:matrix.org](https://matrix.to/#/!unwwlTfOznjEnMMXxf:matrix.org).

## License

* Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>

All files in this repository are licensed under a [GNU Affero General Public
License](LICENCE.md) (version 3 or later).
